/* dazukofs: access control stackable filesystem

   Copyright (C) 2012 Bitdefender
   Copyright (C) 1997-2004 Erez Zadok
   Copyright (C) 2001-2004 Stony Brook University
   Copyright (C) 2004-2007 International Business Machines Corp.
   Copyright (C) 2008-2010 John Ogness
     Author: John Ogness <dazukocode@ogness.net>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __DAZUKOFS_FS_H
#define __DAZUKOFS_FS_H

#define DAZUKOFS_VERSION "3.1.4.20151029"

#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/dcache.h>
#include <linux/mount.h>
#include <linux/rbtree.h>
#include <linux/dcache.h>
#include <linux/spinlock.h>
#include <linux/version.h>

#include "dget.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
extern void *kmem_cache_zalloc(struct kmem_cache *cache, gfp_t flags);
#endif

extern struct kmem_cache *dazukofs_dentry_info_cachep;
extern struct kmem_cache *dazukofs_file_info_cachep;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
extern struct file_operations dazukofs_main_fops;
extern struct file_operations dazukofs_dir_fops;
#else
extern const struct file_operations dazukofs_main_fops;
extern const struct file_operations dazukofs_dir_fops;
#endif
extern struct dentry_operations dazukofs_dops;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
extern struct address_space_operations dazukofs_aops;
#else
extern const struct address_space_operations dazukofs_aops;
#endif

extern int dazukofs_interpose(struct dentry *lower_dentry,
			      struct dentry *dentry, struct super_block *sb,
                              int already_hashed);

struct dazukofs_sb_info {
	struct super_block *lower_sb;
};

struct dazukofs_inode_info {
	struct inode *lower_inode;

	/* the inode (embedded) */
	struct inode vfs_inode;
};

struct dazukofs_dentry_info {
	struct dentry *lower_dentry;
	struct vfsmount *lower_mnt;
};

struct dazukofs_file_info {
	struct file *lower_file;
	int operation;
};

extern struct rb_root d_root;
extern rwlock_t tree_rwlock;
extern struct rb_root dname_cache_root;
extern rwlock_t tree_cache_rwlock;

struct dazukofs_dentry_tree {
	struct rb_node _rb_node;
	struct dentry *_dir;
	int excludeRecursive;
};

struct dazukofs_dname_cache_tree {
	struct rb_node _rb_cache_node;
	unsigned int hash;
	int excludeRecursive;
};

static inline unsigned int RSHash(const char *name)
{
	unsigned int b = 378551;
	unsigned int a = 63689;
	unsigned int hash = 0;

	for (; *name; name++) {
		hash = hash * a + *name;
		a = a * b;
	}
	return hash;
}

static inline void dazukofs_rb_init(void)
{
	d_root = RB_ROOT;
	rwlock_init(&tree_rwlock);

	dname_cache_root = RB_ROOT;
	rwlock_init(&tree_cache_rwlock);
}

static inline void dazukofs_cache_cleanup(void)
{
	struct rb_node *node;

	write_lock(&tree_cache_rwlock);
	node = rb_first(&dname_cache_root);
	while (node) {
		struct dazukofs_dname_cache_tree *entry = rb_entry(node, struct dazukofs_dname_cache_tree, _rb_cache_node);

		rb_erase(node, &dname_cache_root);
		node = rb_first(&dname_cache_root);

		kfree(entry);
	}
	write_unlock(&tree_cache_rwlock);
}

static inline struct dazukofs_dname_cache_tree *_dazukofs_cache_search(const char *name)
{
	unsigned int hash = RSHash(name);
	struct rb_node *node = dname_cache_root.rb_node;
	struct dazukofs_dname_cache_tree *ret = NULL;

	while (node) {
		struct dazukofs_dname_cache_tree *entry = rb_entry(node, struct dazukofs_dname_cache_tree, _rb_cache_node);

		if (entry->hash > hash)
			node = node->rb_left;
		else if (entry->hash < hash)
			node = node->rb_right;
		else {
			ret = entry;
			break;
		}
	}
	return ret;
}

static inline struct dazukofs_dname_cache_tree *dazukofs_cache_search(const char *name)
{
	struct dazukofs_dname_cache_tree *ret;

	read_lock(&tree_cache_rwlock);
	ret = _dazukofs_cache_search(name);
	read_unlock(&tree_cache_rwlock);
	return ret;
}

static inline struct dazukofs_dname_cache_tree *dazukofs_cache_insert(const char *name, int excludeRecursive)
{
	unsigned int value = RSHash(name);
	struct rb_node **link = &dname_cache_root.rb_node;
	struct rb_node *parent = NULL;
	struct dazukofs_dname_cache_tree *new_entry;
	struct dazukofs_dname_cache_tree *ret = NULL;

	write_lock(&tree_cache_rwlock);
	while (*link) {
		struct dazukofs_dname_cache_tree *entry = rb_entry(*link, struct dazukofs_dname_cache_tree, _rb_cache_node);

		parent = *link;
		if (entry->hash > value)
			link = &(*link)->rb_left;
		else if (entry->hash < value)
			link = &(*link)->rb_right;
		else {
			ret = entry;
			goto out;
		}
	}
	new_entry = kmalloc(sizeof(struct dazukofs_dname_cache_tree), GFP_KERNEL);
	if (!new_entry)
		goto out;
	new_entry->hash = value;
	new_entry->excludeRecursive = excludeRecursive;
	rb_link_node(&new_entry->_rb_cache_node, parent, link);
	rb_insert_color(&new_entry->_rb_cache_node, &dname_cache_root);
out:
	write_unlock(&tree_cache_rwlock);
	return ret;
}

static inline void dazukofs_cache_remove_entry(const char *name)
{
	struct dazukofs_dname_cache_tree *old_entry;

	write_lock(&tree_cache_rwlock);
	old_entry = _dazukofs_cache_search(name);
	if (!old_entry) {
		write_unlock(&tree_cache_rwlock);
		return;
	}
	rb_erase(&old_entry->_rb_cache_node, &dname_cache_root);
	write_unlock(&tree_cache_rwlock);
	kfree(old_entry);
}

static inline void dazukofs_rb_cleanup(void)
{
	struct rb_node *node;

	write_lock(&tree_rwlock);
	node = rb_first(&d_root);
	while (node) {
		struct dazukofs_dentry_tree *dir = rb_entry(node, struct dazukofs_dentry_tree, _rb_node);

		rb_erase(node, &d_root);
		node = rb_first(&d_root);

		__dput(dir->_dir);
		kfree(dir);
	}
	d_root = RB_ROOT;
	write_unlock(&tree_rwlock);
}

static inline struct dazukofs_dentry_tree *_dentry_rb_search(struct dentry *value)
{
	struct rb_node *node = d_root.rb_node;

	while (node) {
		struct dazukofs_dentry_tree *dir = rb_entry(node, struct dazukofs_dentry_tree, _rb_node);

		if (dir->_dir > value)
			node = node->rb_left;
		else if (dir->_dir < value)
			node = node->rb_right;
		else
			return dir;
	}
	return NULL;
}

/*
 * We recursively check for exclusion starting with the dentry of current
 * file. We do this until we find a parent that is excluded recursive, or
 * we find root (or mount point - to be decided)
 */
static inline struct dazukofs_dentry_tree *dentry_rb_search(struct dentry *value,
                                                            struct dentry *mount)
{
	struct dazukofs_dentry_tree *retVal = NULL;
	struct dentry *_current = value;
	struct dentry *_previous = NULL;

	read_lock(&tree_rwlock);
	while (_current != _previous) {
		/* looking for root - root->d_parent = root, so no null checks */
		retVal = _dentry_rb_search(_current);
		if (retVal) {
			if (!retVal->excludeRecursive && _current != value) {
				/* go up still */
				retVal = NULL;
				_previous = _current;
				_current = _current->d_parent;
				continue;
			}
			break;
		}
		_previous = _current;
		_current = _current->d_parent;
	}
	read_unlock(&tree_rwlock);
	return retVal;
}

static inline struct dazukofs_dentry_tree *_dentry_rb_insert(struct dazukofs_dentry_tree *new_dir)
{
	struct rb_node **link = &d_root.rb_node;
	struct rb_node *parent = NULL;
	struct dentry *value = new_dir->_dir;

	while (*link) {
		struct dazukofs_dentry_tree *dir = rb_entry(*link, struct dazukofs_dentry_tree, _rb_node);

		parent = *link;
		if (dir->_dir > value)
			link = &(*link)->rb_left;
		else if (dir->_dir < value)
			link = &(*link)->rb_right;
		else
			return dir;
	}
	rb_link_node(&new_dir->_rb_node, parent, link);
	rb_insert_color(&new_dir->_rb_node, &d_root);
	return NULL;
}

static inline int dentry_rb_insert(struct dentry *d_entry, int recursive)
{
	int err = 0;
	struct dazukofs_dentry_tree *new_dir;
	struct dazukofs_dentry_tree *ret;

	new_dir = kmalloc(sizeof(struct dazukofs_dentry_tree), GFP_KERNEL);
	if (!new_dir)
		return -1;
	new_dir->_dir = __dget(d_entry);
	new_dir->excludeRecursive = recursive;
	write_lock(&tree_rwlock);
	ret = _dentry_rb_insert(new_dir);
	write_unlock(&tree_rwlock);
	if (ret) {
		__dput(new_dir->_dir);
		kfree(new_dir);
		err = -1;
	}
	return err;
}

static inline void dentry_rb_remove(struct dentry *old_d_entry)
{
	struct dazukofs_dentry_tree *old_entry;

	write_lock(&tree_rwlock);
	old_entry = _dentry_rb_search(old_d_entry);
	if (!old_entry) {
		write_unlock(&tree_rwlock);
		return;
	}
	rb_erase(&old_entry->_rb_node, &d_root);
	write_unlock(&tree_rwlock);
	__dput(old_entry->_dir);
	kfree(old_entry);
}

static inline struct dazukofs_sb_info *get_sb_private(
						struct super_block *upper_sb)
{
	return upper_sb->s_fs_info;
}

static inline void set_sb_private(struct super_block *upper_sb,
				  struct dazukofs_sb_info *sbi)
{
	upper_sb->s_fs_info = sbi;
}

static inline struct super_block *get_lower_sb(struct super_block *upper_sb)
{
	return get_sb_private(upper_sb)->lower_sb;
}

static inline void set_lower_sb(struct super_block *upper_sb,
				struct super_block *lower_sb)
{
	struct dazukofs_sb_info *sbi = get_sb_private(upper_sb);
	sbi->lower_sb = lower_sb;
}

static inline struct dazukofs_inode_info *get_inode_private(
						struct inode *upper_inode)
{
	return container_of(upper_inode, struct dazukofs_inode_info,
			    vfs_inode);
}

static inline struct inode *get_lower_inode(struct inode *upper_inode)
{
	return get_inode_private(upper_inode)->lower_inode;
}

static inline void set_lower_inode(struct inode *upper_inode,
				   struct inode *lower_inode)
{
	struct dazukofs_inode_info *dii = get_inode_private(upper_inode);
	dii->lower_inode = lower_inode;
}

static inline struct dazukofs_dentry_info *get_dentry_private(
						struct dentry *upper_dentry)
{
	return upper_dentry->d_fsdata;
}

static inline void set_dentry_private(struct dentry *upper_dentry,
				      struct dazukofs_dentry_info *dentryi)
{
	upper_dentry->d_fsdata = dentryi;
}

static inline struct dentry *get_lower_dentry(struct dentry *upper_dentry)
{
	return get_dentry_private(upper_dentry)->lower_dentry;
}

static inline struct vfsmount *get_lower_mnt(struct dentry *upper_dentry)
{
	return get_dentry_private(upper_dentry)->lower_mnt;
}

static inline void set_lower_dentry(struct dentry *upper_dentry,
				    struct dentry *lower_dentry,
				    struct vfsmount *lower_mnt)
{
	struct dazukofs_dentry_info *dii = get_dentry_private(upper_dentry);
	dii->lower_dentry = lower_dentry;
	dii->lower_mnt = lower_mnt;
}

static inline struct dazukofs_file_info *get_file_private(
						struct file *upper_file)
{
	return upper_file->private_data;
}

static inline void set_file_private(struct file *upper_file,
				    struct dazukofs_file_info *filei)
{
	upper_file->private_data = filei;
}

static inline struct file *get_lower_file(struct file *upper_file)
{
	return get_file_private(upper_file)->lower_file;
}

static inline void set_lower_file(struct file *upper_file,
				  struct file *lower_file)
{
	struct dazukofs_file_info *dfi = get_file_private(upper_file);
	dfi->lower_file = lower_file;
}

static inline void set_file_operation_flag(struct file *file, int flag)
{
	if (file->private_data)
		((struct dazukofs_file_info *)(file->private_data))->operation |= flag;
}

static inline int get_file_operation_flag(struct file *file, int flag)
{
	if (!file->private_data)
		return 0;
	return ((struct dazukofs_file_info *)(file->private_data))->operation & flag;
}

#endif  /* __DAZUKOFS_FS_H */
